import datetime
import logging
import os
import tempfile
import traceback

import persistent.dict
import email.Message
from BTrees.Length import Length
from ZODB.POSException import ConflictError
from zope import component
from zope import interface
import zope.interface.common.mapping
import zope.lifecycleevent
import zc.queue.interfaces
import zc.lockfile

import queue

from collective.singing import interfaces

logger = logging.getLogger('collective.singing')

LOCKFILE_NAME = os.path.join(tempfile.gettempdir(),
                             'collective.singing.message.lock')


def dispatch(message):
    dispatcher = interfaces.IDispatch(message.payload)
    try:
        value = dispatcher()
        assert value is not None and len(value) == 2, (
            "Invalid return value from %r: %r" %
            (dispatcher, value))
        status, msg = value
    except ConflictError:
        raise
    except Exception, e:
        status = u'error'
        msg = traceback.format_exc(e)
        logger.exception("Error while dispatching message.")

    message.status_message = msg
    message.status = status
    return message.status, message.status_message


class Message(object):
    interface.implements(interfaces.IMessage)

    _status = None

    def __init__(self, payload, subscription,
                 status=u'new', status_message=None):
        self.payload = payload
        self.subscription = subscription
        self.status_message = status_message
        if status is not None:
            self.status = status

    @apply
    def status():
        def get(self):
            return self._status

        def set(self, value):
            assert value in interfaces.MESSAGE_STATES, value
            old_status = self.status
            self._status = value
            zope.event.notify(MessageChanged(self, old_status))
            self.status_changed = datetime.datetime.now()

        return property(get, set)


class SizedCompositeQueue(zc.queue.CompositeQueue):
    def __init__(self, compositeSize=15,
                 subfactory=zc.queue._queue.BucketQueue):
        super(SizedCompositeQueue, self).__init__(compositeSize, subfactory)
        self.size = 0

    def pull(self, index=0):
        item = super(SizedCompositeQueue, self).pull(index)
        if item:
            self.size -= 1
        return item

    def put(self, item):
        super(SizedCompositeQueue, self).put(item)
        self.size += 1

    def last_item(self):
        # Use instead of [-1]
        # Saves iterating through all items!
        try:
            last_queue = -1
            while not len(self._data[last_queue]):
                last_queue -= 1
            return self._data[last_queue][-1]
        except IndexError:
            return None

    def __len__(self):
        return self.size


class MessageQueues(persistent.dict.PersistentDict):
    interface.implements(interfaces.IMessageQueues)

    def __init__(self, *args, **kwargs):
        super(MessageQueues, self).__init__(*args, **kwargs)
        for status in interfaces.MESSAGE_STATES:
            self[status] = queue.CompositeQueue()
        self._messages_sent = Length()

    @property
    def messages_sent(self):
        return self._messages_sent()

    def dispatch(self):
        try:
            lock = zc.lockfile.LockFile(LOCKFILE_NAME)
        except zc.lockfile.LockError:
            logger.info("Dispatching is locked by another process.")
            return (0, 0)

        try:
            return self._dispatch()
        finally:
            lock.close()

    def _dispatch(self):
        sent = 0
        failed = 0

        for name in 'new', 'retry':
            queue = self[name]
            while True:
                try:
                    message = queue.pull()
                except IndexError:
                    break
                else:
                    status, message = dispatch(message)
                    if status == 'sent':
                        sent += 1
                    else:
                        failed += 1

        self._messages_sent.change(sent)
        return sent, failed

    def clear(self, queue_names=('error', 'sent')):
        for name in queue_names:
            self[name] = self[name].__class__()


class MessageChanged(zope.lifecycleevent.ObjectModifiedEvent):
    interface.implements(interfaces.IMessageChanged)

    def __init__(self, object, old_status, *descriptions):
        super(MessageChanged, self).__init__(object, *descriptions)
        self.old_status = old_status


@component.adapter(interfaces.IMessage, interfaces.IMessageChanged)
def queue_message(message, event):
    # We expect the message to be in *no queue* when we receive this
    # event!
    channel = message.subscription.channel
    queue = channel.queue[message.status]

    # If channel is set to not hold on to sent messages, we log the
    # e-mail message identifier and recipient to allow some minimum
    # record-keeping
    if message.status == 'sent' and not channel.keep_sent_messages:
        if isinstance(message, email.Message.Message):
            message_id = message.payload['Message-ID']
        else:
            message_id = "N/A"

        logger.info("Message sent to \"%s\" <%s> (Message-ID: %s)." % (
            interfaces.ISubscriptionLabel(message.subscription),
            interfaces.ISubscriptionKey(message.subscription),
            message_id
            ))
    else:
        queue.put(message)



import datetime
import logging

import persistent

from zope import interface
from zope import component
from zope.deprecation.deprecation import deprecate

from collective.singing import interfaces
from collective.singing import MessageFactory as _

from plone.memoize import request

logger = logging.getLogger('collective.singing')


class UnicodeFormatter(object):
    interface.implements(interfaces.IFormatItem)

    def __init__(self, item, request):
        self.item = item

    def __call__(self):
        return unicode(self.item)


@request.cache(get_key=lambda fun, obj, req, format: (obj, format))
def format_item(obj, request, format):
    logger.info('Formatting item: (%s, %s, %s)' % (obj, request, format))
    return getIFormatAdapter(obj, request, format)()


def getIFormatAdapter(obj, request, format):
    """Return item formatter for the specified format.

    If a formatter for the specified format is not found, an
    attempt to look up an unspecified formatter is made.
    """

    formatter = component.queryMultiAdapter(
        (obj, request), interfaces.IFormatItem, name=format)

    if formatter is not None:
        return formatter

    return component.getMultiAdapter((obj, request), interfaces.IFormatItem)


class MessageAssemble(object):
    interface.implements(interfaces.IMessageAssemble)
    component.adapts(interfaces.IChannel)

    use_cue = True
    update_cue = True

    def __init__(self, channel):
        self.channel = channel

    def render_message(self, request, subscription,
                       items=(), use_collector=True, override_vars=None):
        channel = self.channel
        collector = channel.collector
        composers = channel.composers
        if override_vars is None:
            override_vars = {}

        subscription_metadata = subscription.metadata
        if subscription_metadata.get('pending'):
            return None

        # Collect items for subscription
        if collector is not None and use_collector:

            # Optionally use and set cue
            if self.use_cue:
                use_cue = subscription_metadata.get(
                    'cue', subscription_metadata.get('date'))
            else:
                use_cue = None

            collector_items, cue = collector.get_items(use_cue, subscription)
            if self.update_cue:
                subscription_metadata['cue'] = cue

            # If there was a collector but no items, we'll skip
            # this subscription:
            if not items and len(collector_items) == 0:
                return None
        else:
            # If there's no collector, we go on with an empty
            # items list:
            collector_items = ()

        raw_items = tuple(items) + tuple(collector_items)

        # First format all items...
        format = subscription_metadata['format']
        formatted_items = [format_item(item, request, format)
                           for item in raw_items]

        # ... then transform them ...
        transformed_items = []
        for formatted_item, raw_item in zip(formatted_items, raw_items):
            for name, transform in component.getAdapters(
                (raw_item,), interfaces.ITransform):
                formatted_item = transform(formatted_item, subscription)
            transformed_items.append(formatted_item)

        # ... and finally render using the right composer. Note
        # that the message is already queued when we render it.
        composer = composers[format]
        return composer.render(subscription, zip(transformed_items, raw_items),
                               override_vars)

    def __call__(self, request, items=(), use_collector=True,
                 override_vars=None):
        if override_vars is None:
            override_vars = {}
        select_optional_collector = override_vars.get("subscriptions_for_collector")
        queued_messages = 0
        for subscription in self.channel.subscriptions.values():
            collector_data = getattr(subscription, "collector_data", {"selected_collectors": set()})
            subscription_selected_collectors = collector_data.get("selected_collectors", set())
            if (select_optional_collector is None or
                  select_optional_collector in subscription_selected_collectors):
              logger.debug("Rendering message for %r." % subscription)
              message = self.render_message(
                  request, subscription, items, use_collector, override_vars)
              if message is not None:
                  queued_messages += 1
        return queued_messages


@deprecate("""\
render_message has become IMessageAssemble.render_message in version
0.6.  Please update your code to use the IMessageAssemble adapter on
IChannel instead.
""")
def render_message(channel, request, subscription, items, use_collector):
    return interfaces.IMessageAssemble(channel).render_message(
        request, subscription, items, use_collector)


@deprecate("""\
assemble_messages has become IMessageAssemble.__call__ in version 0.6.
Please update your code to use the IMessageAssemble adapter on
IChannel instead.
""")
def assemble_messages(channel, request, items=(), use_collector=True):
    return interfaces.IMessageAssemble(channel)(request, items, use_collector)


class AbstractPeriodicScheduler(object):

    """
      >>> from datetime import datetime, date, time, timedelta

      >>> class TestScheduler(AbstractPeriodicScheduler):
      ...     @property
      ...     def now(self):
      ...         return self._now
      ...     def assemble_messages(self, channel, request):
      ...         return self.triggered_last

      >>> def reset(scheduler):
      ...     scheduler.active = True
      ...     scheduler.delta = timedelta(days=7)
      ...     scheduler.triggered_last = datetime(2009, 3, 1, 17, 0)

      >>> scheduler = TestScheduler()
      >>> reset(scheduler)

Delta not reached, triggered_last stays the same.

      >>> last = scheduler.triggered_last
      >>> scheduler._now = datetime(2009, 3, 1, 17, 15)
      >>> scheduler.tick(None, None)
      >>> last == scheduler.triggered_last
      True

      >>> scheduler._now = datetime(2009, 3, 4, 17, 0)
      >>> scheduler.tick(None, None)
      >>> last == scheduler.triggered_last
      True

      >>> scheduler._now = datetime(2009, 3, 8, 16, 59)
      >>> scheduler.tick(None, None)
      >>> last == scheduler.triggered_last
      True

Delta reached.

      >>> scheduler._now = datetime(2009, 3, 8, 17, 15)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 8, 17, 0)

      >>> reset(scheduler)
      >>> scheduler._now = datetime(2009, 3, 8, 17, 0)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 8, 17, 0)

Only the date changes when sending.

      >>> reset(scheduler)
      >>> scheduler._now = datetime(2009, 3, 9, 18, 15)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 8, 17, 0)

      >>> reset(scheduler)
      >>> scheduler._now = datetime(2009, 3, 12, 18, 15)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 8, 17, 0)

If last call is more the delta ago, the interveening sends will
of course  be skipped.

      >>> reset(scheduler)
      >>> scheduler._now = datetime(2009, 3, 26, 18, 15)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 22, 17, 0)

Trigging a manual scheduler (with no delta) always sets it's
triggered_last to now.

      >>> reset(scheduler)
      >>> scheduler.delta = timedelta()
      >>> scheduler._now = datetime(2009, 3, 26, 18, 15)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 26, 18, 15)

      >>> scheduler._now = datetime(2009, 3, 26, 18, 16)
      >>> scheduler.tick(None, None)
      datetime.datetime(2009, 3, 26, 18, 16)

    """
    interface.implements(interfaces.IScheduler)

    triggered_last = datetime.datetime(1970, 1, 1)
    active = False
    delta = None

    def tick(self, channel, request):
        if self.active and (self.now - self.triggered_last >= self.delta):
            return self.trigger(channel, request)

    @property
    def now(self):
        return datetime.datetime.now()

    def trigger(self, channel, request):
        now = self.now

        if self.delta:
            # triggered_last will be set in steps of delta
            while now - self.triggered_last >= self.delta:
                self.triggered_last = self.triggered_last + self.delta
        else:
            self.triggered_last = now

        return self.assemble_messages(channel, request)

    def assemble_messages(self, channel, request):
        return interfaces.IMessageAssemble(channel)(request)

    def __eq__(self, other):
        return (isinstance(other, AbstractPeriodicScheduler) and
                other.delta == self.delta)

    def __ne__(self, other):
        return not self == other

    def __cmp__(self, other):
        """ Don't ask me the exact reason why, but sometimes
        equality between schedulers does not work with only
        __eq__ defined.
        This issue has only been observed under python2.6, and
        is seems to depend on argument order. """
        if not interfaces.IScheduler.providedBy(other):
            return 1
        # TimedScheduler.delta is None
        if self.delta is None or other.delta is None:
            return 1
        return cmp(self.delta, other.delta)


class DailyScheduler(persistent.Persistent, AbstractPeriodicScheduler):
    title = _(u"Daily scheduler")
    delta = datetime.timedelta(days=1)


class WeeklyScheduler(persistent.Persistent, AbstractPeriodicScheduler):
    title = _(u"Weekly scheduler")
    delta = datetime.timedelta(weeks=1)


class ManualScheduler(persistent.Persistent, AbstractPeriodicScheduler):
    title = _(u"Manual scheduler")
    delta = datetime.timedelta()

    def tick(self, channel, request):
        pass


class TimedScheduler(persistent.Persistent, AbstractPeriodicScheduler):
    title = _(u"Timed scheduler")
    active = True
    triggered_last = datetime.datetime(1970, 1, 1)

    def __init__(self):
        super(TimedScheduler, self).__init__()
        self.items = persistent.list.PersistentList()

    def tick(self, channel, request):
        return self.trigger(channel, request, manual=False)

    def trigger(self, channel, request, manual=True):
        count = 0
        now = datetime.datetime.now()
        assembler = interfaces.IMessageAssemble(channel)
        if self.active or manual:
            for when, content, override_vars in tuple(self.items):
                if manual or when < now:
                    self.items.remove((when, content, override_vars))
                    if content is not None:
                        count += assembler(request, (content(),),
                                           override_vars=override_vars)
                    else:
                        count += assembler(request,
                                           override_vars=override_vars)

        # no need to cause a db transaction on every single trigger
        if count > 0:
          self.triggered_last = now

        return count

    def __eq__(self, other):
        return isinstance(other, TimedScheduler)

    def __ne__(self, other):
        return not self == other

schedulers = (WeeklyScheduler, DailyScheduler, ManualScheduler, TimedScheduler)



import unittest
from zope.testing import doctest
from zope.component import testing

from collective.singing import subscribe
import queue
count = 0
root = None


def setUp(test):
    testing.setUp(test)

    # Register adapters and handlers:
    # This query interface relies on a zope.app.catalog to
    # do the job.  Before we can use this catalog, we'll need to register an
    # IIntIds utility and wire in a couple of adatpers defined in the
    # subscribe module.  This is what 'create_subscriptions' does for us:
    from zope.component import provideUtility, provideAdapter, provideHandler
    for adapter in (subscribe.catalog_data,
                    subscribe.SubscriptionSearchableText):
        provideAdapter(adapter)

    from zope.component.event import objectEventNotify
    for handler in (subscribe.subscription_modified,
                    subscribe.subscription_removed,
                    objectEventNotify):
        provideHandler(handler)

    # Set up an IIntIds utility:
    try:
        from zope.intid import IntIds
        from zope.intid.interfaces import IIntIds
        IntIds, IIntIds  # pyflakes
    except ImportError:
        # BBB Plone 4.0 and earlier.
        from zope.app.intid import IntIds
        from zope.app.intid.interfaces import IIntIds
    intids = IntIds()
    provideUtility(intids, IIntIds)

    # We'll register a slight variation of the subscription_added
    # handler that commits the transaction, so that a later lookup of
    # IKeyReference for our subscription will work:
    from ZODB.DemoStorage import DemoStorage
    from ZODB import DB
    global root
    db = DB(DemoStorage())
    root = db.open().root()

    subscription_added.__component_adapts__ = (
        subscribe.subscription_added.__component_adapts__)
    provideHandler(subscription_added)

    # As a last step, we'll register the IKeyReference adapter for all
    # persistent objects:
    from zope.keyreference.persistent import KeyReferenceToPersistent
    from persistent.interfaces import IPersistent
    provideAdapter(KeyReferenceToPersistent, adapts=(IPersistent,))

    provideAdapter(subscribe.get_subscription_label)
    provideAdapter(subscribe.get_subscription_key)


def subscription_added(obj, event):
    global count
    count += 1
    root[str(count)] = obj
    from transaction import commit
    commit()
    subscribe.subscription_added(obj, event)


def test_suite():
    return unittest.TestSuite([

        doctest.DocFileSuite('README.txt'),

        doctest.DocFileSuite(
            'async.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
        ),

        doctest.DocFileSuite(
            'collector.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
        ),

        doctest.DocFileSuite(
           'message.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocFileSuite(
            'scheduler.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
        ),

        doctest.DocFileSuite(
            'subscribe.txt',
            setUp=setUp, tearDown=testing.tearDown,
        ),

        doctest.DocFileSuite(
            'browser/converters.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
        ),

        doctest.DocFileSuite(
            'queue.txt',
            globs={'Queue':lambda: queue.CompositeQueue(2)}
        ),

       doctest.DocTestSuite(
           'collective.singing.mail',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'collective.singing.subscribe',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'collective.singing.scheduler',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        ])



from zope import interface
from zope import schema
import zope.component.interfaces
from zope.interface.interfaces import IInterface
from zope.interface.common.mapping import IMapping
from zope.annotation.interfaces import IAnnotatable
import z3c.form.interfaces

from collective.singing import MessageFactory as _


class ICollector(interface.Interface):
    """Collectors are useful for automatic newsletters.  They are
    responsible for assembling a list of items for publishing.
    """

    title = schema.TextLine(
        title=_(u"Title"),
        )

    optional = schema.Bool(
        title=_(u"Subscriber optional"),
        )

    significant = schema.Bool(
        title=_(u"Significant"),
        description=_(u"Include items from this collector even if there are "
                      u"no items returned by significant siblings.")
        )

    schema = schema.Object(
        title=_(u"A schema instance for use in the subscription form"),
        description=_(u"Values are stored via the ICollectorData adapter per "
        "subscriber."),
        required=False,
        schema=IInterface,
        )

    def get_items(cue=None, subscription=None):
        """Return a tuple '(items, cue)' where 'items' is the items
        that'll be fed into the composer, and 'cue' is an object that
        can be passed to this method again.

        If 'cue' is passed, I will only return items that are newer
        than the items that were returned when the cue was retrieved.

        If 'subscription' is given, I will filter according to the
        subscriber's ICollectorData.
        """


class IComposer(interface.Interface):
    """Composers will typically provide a user interface that lets you
    modify the look of the message rendered through it.
    """

    name = schema.TextLine(
        title=_(u"The Composer's format, e.g. 'html'"),
        )

    title = schema.TextLine(
        title=_(u"The Composer's title, e.g. 'HTML E-Mail'"),
        )

    schema = schema.Object(
        title=_(u"A schema instance for use in the subscription form"),
        description=_(u"Values are stored via the IComposerData adapter per "
        "subscriber."),
        schema=IInterface,
        )

    def render(subscription, items=(), override_vars=None):
        """Given a subscription and a list of items, I will create an
        IMessage and return it.

        The ``items`` argument is a list of 2-tuples of the form
        ``(formatted, original)``, where ``original`` is the item as
        it was retrieved from the collector, and ``formatted`` is the
        result of running the item through all applicable formatters
        and transforms.  Making use of the ``original`` item will
        obviously bind the implementation of the composer to that of
        the collector.  However, it's considered useful for custom
        implementations that need total control and that know what
        collector they'll be using.
        """

    def render_confirmation(subscription):
        """Given a subscription, I will create an IMessage that the
        user has to react to in order to confirm the subscription, and
        return it.
        """

    def render_forgot_secret(subscription):
        """Given a subscription, I will create an IMessage that links
        the user to his subscription page.
        """


class IScheduler(interface.Interface):
    """A scheduler triggers the sending of messages periodically.
    """
    title = schema.TextLine(
        title=_(u"Title"),
        )

    triggered_last = schema.Datetime(
        title=_(u"Triggered the last time"),
        )

    active = schema.Bool(
        title=_(u"Active"),
        )

    def tick(channel, request):
        """Check if messages need to be assembled and sent; return the
        number of messages queued or None.

        This method is guaranteed to be called periodically.
        """

    def trigger(channel, request):
        """Assemble and queue messages; return the number of messages
        queued.

        A manual override.
        """


class ISubscriptions(interface.Interface):
    """The `subscriptions` attribute of IChannel implements this.

    Allows for adding, removing, and querying subscriptions.
    """
    def values():
        """Iterate over all subscription objects.
        """

    def add_subscription(
        channel, secret, composer_data, collector_data, metadata):
        """Add a subscription and return it.

        Raises ValueError if subscription already exists.
        """

    def remove_subscription(subscription):
        """Remove subscription.
        """

    def query(**kwargs):
        """Search subscriptions.

        Available fields are: 'fulltext', 'secret', 'format', 'key', 'label'.
        """


class ISubscription(IAnnotatable):
    """A subscription to a channel.
    """

    channel = schema.Object(
        title=_(u"The channel that we're subscribed to"),
        schema=IInterface,  # should be really IChannel
        )

    secret = schema.ASCIILine(
        title=_(u"The subscriber's secret, ideally unique across channels"),
        description=u"""\
        Might be a hash of the subscriber's e-mail and a secret on the
        server in case we're dealing with an anonymous subscription.
        Used for unsubscription or for providing an overview of all of
        a subscriber's subscriptions.
        """,
        )

    composer_data = schema.Dict(title=_(u"Composer data"))
    collector_data = schema.Dict(title=_(u"Collector data"))
    metadata = schema.Dict(title=_(u"Metadata"))


class ISubscriptionLabel(interface.Interface):
    """Marker interface for field on IComposerSchema that labels a
    subscription.
    """


class ISubscriptionKey(interface.Interface):
    """Marker interface for field on IComposerSchema that serves as a
    key for conflicts.

    For subscriptions per e-mail this is the e-mail field.  No
    subscriptions for the same format and same e-mail address will be
    allowed, then.
    """


class IMessageQueues(IMapping):
    """A dict that contains one ``zc.queue.interfaces.IQueue`` per
    message status.
    """
    messages_sent = schema.Int(
        title=_(u"Total number of messages sent through this queue"),
        )


class IChannel(interface.Interface):
    """A Channel is what we can subscribe to.

    A Channel is a hub of configuration.  It doesn't do anything by
    itself.  Rather, it provides a number of components to configure
    and work with.  It is also the container of subscriptions to it.
    """

    name = schema.ASCIILine(
        title=_(u"Unique identifier for this channel across the site."),
        )

    title = schema.TextLine(
        title=_(u"Title"),
        )

    description = schema.Text(
        title=_(u"Description")
        )

    subscribeable = schema.Bool(
        title=_(u"Subscribeable"),
        default=False
        )

    scheduler = schema.Object(
        title=_(u"Scheduler (when)"),
        required=False,
        schema=IScheduler,
        )

    collector = schema.Object(
        title=_(u"Collector (what)"),
        required=False,
        schema=ICollector,
        )

    composers = schema.Dict(
        title=_(u"The channel's composers, keyed by format."),
        key_type=schema.TextLine(title=_(u"Format")),
        value_type=schema.Object(title=_(u"Composer"), schema=IComposer),
        )

    subscriptions = schema.Object(
        title=_(u"The channel's subscriptions"),
        schema=ISubscriptions,
        )

    queue = schema.Object(
        title=_(u"This channel's message queues, keyed by message status"),
        schema=IMessageQueues,
        )

    keep_sent_messages = schema.Bool(
        title=_(u"Keep a record of sent messages."),
        description=_(u"This is not currently recommended for large volumes "
                      u"of messages due to storage requirements."),
        default=False,
        )


class IMessageAssemble(interface.Interface):
    """An adapter on the channel that's usually invoked by a scheduler
    to render and queue new messages.
    """
    def __call__(request, items=(), use_collector=True, override_vars=None):
        """Process all subscribers.

          o ``items`` is the items added manually

          o ``use_collector`` is a boolean indicating whether the
            channel's collector should be considered at all

        Returns the number of queued messages."""

    def render_message(request, subscription, items=(), use_collector=True,
                       override_vars=None):
        """Render and queue messages for an individual subscription.

          o ``subscription`` is an ISubscription

        For other arguments, please see the docstring of ``__call__``.

        Returns the IMessage if a message was queued or None to
        indicate that no message was queued.
        """


class IChannelLookup(interface.Interface):
    """A utility that looks up all channels in a site.
    """

    def __call__():
        """Return a list of ``IChannel`` objects.
        """


class ISalt(interface.Interface):
    """A utility that's a salt for use in creating secrets.
    """


class IRequestBasedSecret(interface.Interface):
    """A utility that provides a secret based on the request.
    """
    def __call__(request):
        """Return an ASCII secret.
        """


class IComposerBasedSecret(interface.Interface):
    """An adapter for composer that provides a secret based on the
    user's data used for the composer (see also 'IComposerData').
    """

    def secret(data):
        """Return an ASCII secret.

        'data' is a dict that corresponds to the IComposer's schema.
        """


MESSAGE_STATES = [u'new',    # just added
                  u'sent',   # sent successfully
                  u'error',  # error while sending
                  u'retry',  # error while sending, but retrying
                  ]


class IMessage(interface.Interface):
    """Messages are objects ready for sending.
    """
    payload = schema.Field(
        title=_(u"The message's payload, e.g. the e-mail message."),
        )

    subscription = schema.Object(
        title=_(u"Subscription, referenced for bookkeeping purposes only."),
        schema=ISubscription,
        )

    status = schema.Choice(
        title=_(u"State"),
        description=_(u"IMessageChanged is fired automatically when this "
                      u"is set"),
        values=MESSAGE_STATES)

    status_message = schema.Text(
        title=_(u"Status details"),
        required=False)

    status_changed = schema.Datetime(
        title=_(u"Last time this message changed its status"),
        )


class IMessageChanged(zope.lifecycleevent.interfaces.IObjectModifiedEvent):
    """An object event on the message that signals that the status has
    changed.
    """
    old_status = schema.TextLine(title=_(u"Old status of message"))


class IDispatch(interface.Interface):
    """Dispatchers adapt message *payloads* and send them."""

    def __call__():
        """Attempt to send message.

        Must return a tuple ``(status, status_message)``.  See
        ``MESSAGE_STATES`` for possible choices for the status.  The
        status message may be None or a text containing details about
        the status, e.g. why it failed.

        If this method raises an exception, an 'error' is assumed.
        """


class IFormatItem(interface.Interface):
    """A view that formats an item for use in a newsletter.

    Given an item as retrieved from a collector, this view returns a
    representation of the given item ready for inclusion in the
    message via the ``IComposer.render`` method.
    """

    def __call__():
        """Returns a unicode-string."""


class ITransform(interface.Interface):
    """An adapter that allows to rewrite links and the like in
    outgoing messages.
    """
    def __call__(text, subscription):
        """Return transformed text
        """


class IFormLayer(z3c.form.interfaces.IFormLayer):
    pass


class IDynamicVocabularyCollection(zope.schema.interfaces.ICollection):
    """Used to mark ICollection Fields in our dynamic schemas.  To
    indicate that thay have dynamic vocabularies.  We need a more
    tolerant z3c.form.interfaces.IDataConverter in this case - one
    that drops saved values that are no longer in the vocabulary when
    updating widgets.

    (XXX: Mabye this should go into the browser package?)
    """


class ICollectorSchema(interface.Interface):
    """An adapter from subscription to the collector data.
    """



import os
from setuptools import find_packages
from setuptools import setup

version = '0.7.4'


def read(*pathnames):
    return open(os.path.join(os.path.dirname(__file__), *pathnames)).read()

setup(
    name='collective.singing',
    version=version,
    description="A Zope 3 library for sending notifications and newsletters",
    long_description='\n'.join([
        read('docs', 'README.txt'),
        read('docs', 'HISTORY.txt'),
    ]),
    classifiers=[
        "Framework :: Plone",
        "Framework :: Plone :: 3.2",
        "Framework :: Plone :: 3.3",
        "Framework :: Plone :: 4.0",
        "Framework :: Plone :: 4.1",
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='zope plone notification newsletter',
    author='Daniel Nouri, Thomas Clement Mogensen and contributors',
    author_email='singing-dancing@googlegroups.com',
    url='http://pypi.python.org/pypi/collective.singing',
    license='GPL',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['collective'],
    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'setuptools',
        'zope.sendmail',
        'zope.keyreference>=3.6.2',
        'zope.app.keyreference==3.6.1',
        'zope.app.catalog',
        'zope.app.i18n',
        'zc.queue<2.0.0a1',
        'zc.lockfile',
        'plone.z3cform>=0.5.1',
    ],

    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



import unittest
from zope.testing import doctest

from zope.component import testing
import z3c.form.testing
import plone.z3cform.tests

from collective.singing.tests import setUp


def setup_defaults():
    # Set up z3c.form defaults
    z3c.form.testing.setupFormDefaults()

    # And plone.z3cform
    plone.z3cform.tests.setup_defaults()


def test_suite():
    return unittest.TestSuite([

        doctest.DocFileSuite(
           'wizard.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocFileSuite(
           'subscribe.txt',
           setUp=setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'collective.singing.browser.subscribe',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        ])
